/*
    Copyright (C) 2002 Steve Harris <steve@plugin.org.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>

#include "ladspa-util.h"

#define BLO_N_WAVES 4
#define BLO_SINE    0
#define BLO_TRI     1
#define BLO_SQUARE  2
#define BLO_SAW     3

#ifndef BLO_N_HARMONICS
#define BLO_N_HARMONICS 64
#endif

#define BLO_SIN_GEN(phase) sin((phase) * 2.0f * (float)M_PI / table_size_f)
#define BLO_NEXT_TABLE     (all_tables + (table_count++ * table_size))

typedef struct {
	float *h_tables[BLO_N_WAVES][BLO_N_HARMONICS];
	int table_size;
	int table_mask;
} blo_h_tables;

typedef union {
	int all;
	struct {
		unsigned short fr;
		short in;
	} part;
} blo_fixp;

typedef struct {
	blo_h_tables *tables;
	float sample_rate;
	float nyquist;
	unsigned int wave;
	blo_fixp ph;
	blo_fixp om;
	float ph_coef;
	int ph_mask;
	int table_mask;
	int table_size;
	int topbit;
	float *table;
	float *table_b;
	float xfade;
} blo_h_osc;

blo_h_tables *blo_h_tables_new(int table_size);

blo_h_osc *blo_h_new(blo_h_tables *tables, unsigned int wave, float sample_rate);

/* Set frequency for static oscilator, less cycles, but won't modulate as well
 * You can't use blo_osc_hd_run_* until you've called the _hd version of this
 * function.
 */

static inline void blo_hs_set_freq(blo_h_osc *this, const float f)
{
	unsigned int tab_num;
	const float ff = fabs(f);

	// This needs to be a cast, no idea why
	this->om.all = (int)(f * this->ph_coef);
	tab_num = f_round(this->nyquist / ff - 0.5f);
	if (tab_num >= BLO_N_HARMONICS) {
		tab_num = BLO_N_HARMONICS - 1;
	}
	this->table_b = this->tables->h_tables[this->wave][tab_num];
}

/* Set frequency for dynamic oscilator, can only used with _hd run calls.
 */

static inline void blo_hd_set_freq(blo_h_osc *this, const float f)
{
	unsigned int tab_num;
	const float ff = fabs(f);

	// This needs to be a cast, no idea why
	this->om.all = (int)(f * this->ph_coef);
	tab_num = f_round(this->nyquist / ff - 0.5f);
	if (tab_num >= BLO_N_HARMONICS) {
		tab_num = BLO_N_HARMONICS - 1;
	}
	this->table = this->tables->h_tables[this->wave][tab_num];
	this->xfade = this->nyquist / ff - tab_num;
	if (this->xfade > 1.0f) {
		this->xfade = 1.0f;
	}
	if (tab_num > 0) {
		tab_num--;
	}
	this->table_b = this->tables->h_tables[this->wave][tab_num];
}

/* Run static oscilator, returns amplitude for current phase and advances the
 * phase. Uses linear interpoation */

static inline float blo_hs_run_lin(blo_h_osc *this)
{
	const float frac = (float)(this->ph.part.fr) * 0.00001525878f;
	const int idx = this->ph.part.in;

	this->ph.all += this->om.all;
	this->ph.all &= this->ph_mask;
	if (this->topbit != (this->ph.all & this->table_size)) {
		this->topbit = this->ph.all & this->table_size;
		this->table = this->table_b;
	}

	return this->table[idx] * (1.0f - frac) + this->table[(idx+1) & this->table_mask] * frac;
}

/* Run static oscilator, returns amplitude for current phase and advances the
 * phase. Uses cubic interpoation */

static inline float blo_hs_run_cub(blo_h_osc *this)
{
	const float frac = (float)(this->ph.part.fr) * 0.00001525878f;
	const int idx = this->ph.part.in;
	float *t = this->table;
	const int tm = this->table_mask;

	this->ph.all += this->om.all;
	this->ph.all &= this->ph_mask;

	if (this->topbit != (this->ph.all & this->table_size)) {
		this->topbit = this->ph.all & this->table_size;
		this->table = this->table_b;
		t = this->table_b;
	}

	return cube_interp(frac, t[(idx-1) & tm], t[idx], t[(idx+1) & tm], t[(idx+2) & tm]);
}

/* Run dynamic oscilator, returns amplitude for current phase and advances the
 * phase, ensures harmonics won't suddently pop into existence, takes more
 * cycles and has slightly less high frequency partials than the static
 * version */

static inline float blo_hd_run_lin(blo_h_osc * const this)
{
	float low, high;
	const float frac = (float)(this->ph.part.fr) * 0.00001525878f;
	const int idx = this->ph.part.in;

	this->ph.all += this->om.all;
	this->ph.all &= this->ph_mask;

	low = this->table_b[idx] * (1.0f - frac) + this->table_b[(idx+1) & this->tables->table_mask] * frac;
	high = this->table[idx] * (1.0f - frac) + this->table[(idx+1) & this->table_mask] * frac;

	return low * (1.0f - this->xfade) + high * this->xfade;
}

/* Run dynamic oscilator, returns amplitude for current phase and advances the
 * phase, ensures harmonics won't suddently pop into existence, takes more
 * cycles and has slightly less high frequency partials than the static
 * version. This one uses cubic interpolation. */

static inline float blo_hd_run_cub(blo_h_osc * const this)
{
	float low, high;
	const float frac = (float)(this->ph.part.fr) * 0.00001525878f;
	const int idx = this->ph.part.in;
	const float *tl = this->table_b;
	const float *th = this->table;
	const int tm = this->table_mask;


	this->ph.all += this->om.all;
	this->ph.all &= this->ph_mask;

	low = cube_interp(frac, tl[(idx-1) & tm], tl[idx], tl[(idx+1) & tm], tl[(idx+2) & tm]);
	high = cube_interp(frac, th[(idx-1) & tm], th[idx], th[(idx+1) & tm], th[(idx+2) & tm]);

	return low * (1.0f - this->xfade) + high * this->xfade;
}
