/* Some misc util functions for audio DSP work, written by Steve Harris,
   December 2000

   steve@plugin.org.uk
*/

#ifndef LADSPA_UTIL_H
#define LADSPA_UTIL_H

#include <math.h>

#include "config.h"

// Sometimes it doesn't get defined, even though it eists and C99 is declared
long int lrintf (float x);

// 1.0 / ln(2)
#define LN2R 1.442695041f

// Denormalise floats, only actually needed for PIII and recent PowerPC
#define FLUSH_TO_ZERO(fv) (((*(unsigned int*)&(fv))&0x7f800000)==0)?0.0f:(fv)

// Limit a value to be l<=v<=u
#define LIMIT(v,l,u) ((v)<(l)?(l):((v)>(u)?(u):(v)))

// Truncate-to-zero modulo (ANSI C doesn't specify) will only work
// if v > -m and v < 2m
#define MOD(v,m) (v<0?v+m:(v>=m?v-m:v))

// Truncate-to-zero modulo (ANSI C doesn't specify) will only work
// if v > -m and v < m
#define NEG_MOD(v,m) ((v)<0?((v)+(m)):(v))

// Convert a value in dB's to a coefficent
#define DB_CO(g) ((g) > -90.0f ? powf(10.0f, (g) * 0.05f) : 0.0f)
#define CO_DB(v) (20.0f * log10f(v))

// Linearly interpolate
#define LIN_INTERP(f,a,b) ((a) + (f) * ((b) - (a)))

// Cubic interpolation function
static inline float cube_interp(const float fr, const float inm1, const float
                                in, const float inp1, const float inp2)
{
	return in + 0.5f * fr * (inp1 - inm1 +
	 fr * (4.0f * inp1 + 2.0f * inm1 - 5.0f * in - inp2 +
	 fr * (3.0f * (in - inp1) - inm1 + inp2)));
}

// Truncate float to int
static inline int f_trunc(float f) {
        f -= 0.5f;
        f += (3<<22);
        return *((int*)&f) - 0x4b400000;
}

#ifdef HAVE_LRINTF

#define f_round(f) lrintf(f)

#else

// Round float to int (faster than f_trunc)
static inline int f_round(float f) {
        f += (3<<22);
        return *((int*)&f) - 0x4b400000;
}

#endif

// 16.16 fixpoint
typedef union {
	int all;
	struct {
		unsigned short fr;
		short in;
	} part;
} fixp16;

// 32.32 fixpoint
typedef union {
	long long all;
	struct {
		unsigned int fr;
		int in;
	} part;
} fixp32;


/* Andrew Simper's pow(2, x) aproximation from the music-dsp list */
static inline float f_pow2(float x)
{
	long *px = (long*)(&x); // store address of float as long pointer
	const float tx = (x-0.5f) + (3<<22); // temporary value for truncation
	const long  lx = *((long*)&tx) - 0x4b400000; // integer power of 2
	const float dx = x-(float)(lx); // float remainder of power of 2

	x = 1.0f + dx*(0.6960656421638072f + // cubic apporoximation of 2^x
                   dx*(0.224494337302845f +  // for x in the range [0, 1]
                   dx*(0.07944023841053369f)));
	*px += (lx<<23); // add integer power of 2 to exponent

	return x;
}

/* Fast exponentiation function, y = e^x */
#define f_exp(x) f_pow2(x * LN2R)

#endif
